/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2009 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 *
 *******************************************************************************
 *
 * $Id$
 *
 ******************************************************************************/
package it.tidalwave.metadata.viewer.test;

import it.tidalwave.metadata.viewer.*;
import java.lang.reflect.Method;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyDescriptor;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.util.Collections;
import java.util.Comparator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import it.tidalwave.util.logging.Logger;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.awt.BorderLayout;
import java.awt.Component;
import javax.swing.ComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.swing.JComponent;
import org.jdesktop.beansbinding.BeanProperty;
import org.jdesktop.beansbinding.Property;
import it.tidalwave.beans.FastBeanProperty;
import it.tidalwave.metadata.spi.MetadataItemEnhancer;
import java.io.InputStream;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import javax.annotation.Nonnull;
import it.tidalwave.util.logging.SingleLineLogFormatter;
import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import static org.junit.Assert.*;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public abstract class PaneTestSupport<P extends MetadataItemPaneSupport<Bean>, Bean> 
  {
    private static final String CLASS = PaneTestSupport.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private static final List<String> IGNORED_PROPS = Arrays.asList("ancestor", "margin", "Frame.active");
    
    // Define this property in nbproject/private/project.propeties as test-unit-sys-prop.ui.test.delay
    public static final int FRAME_VISIBILITY_DELAY = Integer.parseInt(System.getProperty("ui.test.delay", "10"));
    
    private JFrame frame;
    private Class<P> paneClass;
    private Class<Bean> beanClass;
    protected P pane;
    
    /***************************************************************************
     *
     * This listener asserts that notified changes are performed within the
     * Event Dispatch Thread.
     *
     **************************************************************************/
    class EventDispatchThreadAssertor implements PropertyChangeListener, VetoableChangeListener
      {
        public String latestPropertyChanged;
        
        public void propertyChange (final PropertyChangeEvent event) 
          {
            assertChangeInEDT(event);
          }
        
        public void vetoableChange (final PropertyChangeEvent event) 
          throws PropertyVetoException 
          {
            assertChangeInEDT(event);
          }
        
        private void assertChangeInEDT (final PropertyChangeEvent event)
          {
            final String propertyName = event.getPropertyName();
            
            if (!IGNORED_PROPS.contains(propertyName))
              {
                final Object source = event.getSource();
                final String componentName = (source instanceof Component) ? ((Component)source).getName() : source.toString();
                final String message = String.format("!! Property '%s.%s' changed in a thread which is not EDT! %s->%s",
                                                     componentName, propertyName,
                                                     event.getOldValue(), event.getNewValue());
                assertTrue(message, SwingUtilities.isEventDispatchThread()); // FIXME
                logger.info("Property '%s.%s' correctly changed in the EDT: %s->%s", componentName, propertyName,
                                                                                     event.getOldValue(), event.getNewValue());
                latestPropertyChanged = propertyName;
              }
          }
      }
    
    private final EventDispatchThreadAssertor eventDispatchThreadAssertor = new EventDispatchThreadAssertor();

    /***************************************************************************
     *
     *
     **************************************************************************/
    protected PaneTestSupport (final Class<Bean> beanClass, final Class<P> paneClass)
      {
        this.beanClass = beanClass;
        this.paneClass = paneClass;    
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @BeforeClass
    public static void setupLogging()
      throws Exception
      {
        try
          {
            final InputStream is = PaneTestSupport.class.getResourceAsStream("log.properties");
            LogManager.getLogManager().readConfiguration(is);
            is.close();

            //
            // The formatter must be set programmatically as the property in the log.properties won't
            // be honored. I suspect it is related with NetBeans module classloaders as the formatter
            // is loaded inside LogManager by using the SystemClassLoader, which only sees the classpath.
            //
            final SingleLineLogFormatter formatter = new SingleLineLogFormatter();
            java.util.logging.Logger rootLogger = java.util.logging.Logger.getLogger(PaneTestSupport.class.getName());

            while (rootLogger.getParent() != null)
              {
                rootLogger = rootLogger.getParent();
              }

            for (final Handler handler : rootLogger.getHandlers())
              {
                handler.setFormatter(formatter);
              }
          }
        catch (Exception e)
          {
            e.printStackTrace();
          }
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    @Before
    public void setupPane()
      throws Exception
      {
        if (paneClass != null)
          {
            pane = paneClass.newInstance();
            
            SwingUtilities.invokeAndWait(new Runnable() 
              {
                public void run() 
                  {
                    frame = new JFrame("Test for " + paneClass.getName());
                    frame.getContentPane().add(pane, BorderLayout.CENTER);
                    frame.pack();
                    frame.setVisible(true);
                    addEventDispatchThreadAssertor(pane, eventDispatchThreadAssertor);
                  }
              });  
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    @After
    public void disposePane()
      throws InterruptedException 
      {
        if (frame != null)
          {
            final long baseTime = System.currentTimeMillis();

            while (frame.isVisible() && (System.currentTimeMillis() - baseTime < FRAME_VISIBILITY_DELAY))
              {
                Thread.sleep(100);
              }

            frame.dispose();
          }
      }

    /***************************************************************************
     * 
     * Tests if populating the pane works (this means to test if binding 
     * synchronization works from the bean to the pane). The pane is first 
     * filled with the
     * given contents by calling {@link MetadataItemPaneSupport#bind(Object)},
     * then its contents are re-read and compared with the reference.
     * 
     * @param  contents  the stuff to populate the pane with
     *
     **************************************************************************/
    protected final void testPopulate (final Bean contents)
      throws Exception
      {
        logger.info("===== %s testPopulate(%s)", getClass(), contents);
        final Bean bean = createDecoratorBean(clone(contents));
        pane.bind(new MetadataItemHolderMock<Bean>(bean));
        assertPaneContents(contents);
      }
    
    /***************************************************************************
     *
     * Tests whether property changes in the bound bean are correctly propagated
     * to the pane (and in the EDT thread). The test first binds some
     * initial contents to the pane, then changes one by one its properties
     * according to a second bean.
     * It is possible to specify a set of derived properties - they are bound
     * to other properties and change in sync, so they should be ignored in the
     * test.
     *
     * @param  initialContents    the initial contents
     * @param  newContents        the new contents
     * @param  derivedProperties  the ignored properties
     *
     **************************************************************************/
    protected final void testChange (@Nonnull final Bean initialContents,
                                     @Nonnull final Bean newContents,
                                     @Nonnull final String ... derivedProperties)
      throws Exception
      {
        logger.info("===== %s testChange(%s, %s)", getClass(), initialContents, newContents);
        assertFalse(SwingUtilities.isEventDispatchThread());
        final Bean bean = createDecoratorBean(clone(initialContents));
        pane.bind(new MetadataItemHolderMock<Bean>(bean));
        addPropertyChangeListener(pane.getEDTBean(), eventDispatchThreadAssertor);
        assertPaneContents(initialContents);
        
        final PropertyDescriptor[] descriptors = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
        
        for (final PropertyDescriptor descriptor : sorted(Arrays.asList(descriptors)))
          {
            final String propertyName = descriptor.getName();
            
            // TODO: can't test derivative properties since they would set the same value as the
            // base property and this won't trigger a change.
            if (!propertyName.endsWith("Available")
                && !propertyName.endsWith("AsDate")  // *AsDate is a fix for EXIF
                && !propertyName.equals("userCommentAsString")
                && !Arrays.asList(derivedProperties).contains(propertyName))
              {
                final Property property = FastBeanProperty.create(propertyName);

                if (property.isReadable(newContents) && property.isWriteable(bean))
                  {
                    logger.info(">>>> Changing '%s' from outside the EventDispatchThread...", propertyName);
                    eventDispatchThreadAssertor.latestPropertyChanged = null;
                    property.setValue(bean, property.getValue(newContents)); 
                    // FIXME: this fails, the event is not fired
                    assertNotNull(String.format("Change not detected for %s", propertyName),
                                  eventDispatchThreadAssertor.latestPropertyChanged);
                  }
              }
          }
        
        assertPaneContents(newContents);
      }
    
    /***************************************************************************
     *
     * Tests if capturing the user's input works (this means to test if binding
     * synchronization works from the pane to the bean). The pane is filled by
     * emulating typing into it, then the values in the bound bean are compared
     * with the reference.
     *
     * @param  input  the stuff to type into the pane
     *
     **************************************************************************/
    protected final void testInput (final Bean input)
      throws Exception
      {
        logger.info("===== %s testInput(%s)", getClass(), input);
        final Bean result = createDecoratorBean(beanClass.newInstance());
        pane.bind(new MetadataItemHolderMock<Bean>(result));
        
        SwingUtilities.invokeAndWait(new Runnable() 
          {
            public void run() 
              {
                logger.fine(">>>> starting fillInPane(), getting data from %s...", input);
                fillInPane(input);
                logger.fine(">>>> fillInPane() completed");
              }
          });
        
        dump(result);
        assertBean(input, result);
      }
    
    /***************************************************************************
     *
     * Asserts that two beans are equal, property by property. While calling
     * the {@link Object#equals(Object)} would do (supposing that it is 
     * correctly implemented for the bean), it would not provide good 
     * diagnostics about what properties differ, so each test must concretely
     * implement this by explicit property-by-property comparisons.
     * 
     * TODO: use Introspection to iterate through properties and test them,
     * then make this method final.
     * 
     * @param  expected  the bean with expected values
     * @param  actual    the bean to test
     * 
     **************************************************************************/
    protected abstract void assertBean (final Bean expected, final Bean actual);

    /***************************************************************************
     *
     * Asserts that the pane is filled with expected values.
     * 
     * @param  expected  the bean with expected values
     *
     **************************************************************************/
    protected abstract void assertPaneContents (final Bean expected);
    
    /***************************************************************************
     *
     * Fills in the pane with the given values.
     * 
     * @param  reference  the values to fill the pane with
     *
     **************************************************************************/
    protected abstract void fillInPane (final Bean reference);
    
    protected List<String> getPropertiesInConstrainedOrder()
      {
        return Collections.emptyList(); 
      }
    
    protected final List<PropertyDescriptor> sorted (final List<PropertyDescriptor> propertyDescriptors)
      {
        final List<String> specialOrder = getPropertiesInConstrainedOrder();
        
        Collections.sort(propertyDescriptors, new Comparator<PropertyDescriptor>() 
          {
            public int compare (final PropertyDescriptor p1, final PropertyDescriptor p2) 
              {
                String n1 = p1.getName();
                String n2 = p2.getName();
                
                final int i1 = specialOrder.indexOf(n1);
                final int i2 = specialOrder.indexOf(n2);
                
                if (i1 >= 0)
                  {
                    n1 = Integer.toString(i1) + n1;  
                  }
                
                if (i2 >= 0)
                  {
                    n2 = Integer.toString(i2) + n2;  
                  }
                
                return n1.compareTo(n2);
              }
          });  
   
        return propertyDescriptors;    
      }
            
    /***************************************************************************
     *
     * Clones a bean. This method doesn't call {@link Object#clone()}, but
     * uses serialization in a memory buffer.
     * 
     * @param  bean  the bean to clone
     * @return       the cloned bean
     *
     **************************************************************************/
    protected static <T> T clone (final T bean)
      throws Exception
      {
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(bean);
        oos.close();
        final ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
        final ObjectInputStream ois = new ObjectInputStream(bais);
        final T result = (T)ois.readObject();
        ois.close();
        
        return result;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected static Collection<String> fromCommaSeparatedString (final String string)
      {
        final List<String> result = new ArrayList<String>();
        final String[] split = string.split(",");
        
        for (final String s : split)
          {
            result.add(s.trim());
          } 
        
        return result;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected static String toCommaSeparatedString (final Collection<String> collection)
      {
        final StringBuilder builder = new StringBuilder();
        
        for (final String s : collection)
          {
            if (builder.length() > 0)
              {
                builder.append(", ");  
              }
            
            builder.append(s);
          }
        
        return builder.toString();
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected static void assertEqualsSelections (final Collection<?> expected, final Collection<?> value)
      {
        final Collection<?> temp = new ArrayList<Object>(value);
        temp.removeAll(expected);
        assertEquals(1, temp.size());
        assertEquals(null, temp.iterator().next());
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected static Collection<Object> findElements (final JComboBox comboBox)
      {
        final Collection<Object> result = new ArrayList<Object>();
        final ComboBoxModel model = comboBox.getModel();
        
        for (int i = 0; i < model.getSize(); i++)
          {
            result.add(model.getElementAt(i));
          }
        
        return result;
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    protected static void dump (final Object bean) 
      throws IntrospectionException
      {
        logger.info("======= dump(%s)", bean);
        final PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(bean.getClass()).getPropertyDescriptors();
        
        for (final PropertyDescriptor descriptor : propertyDescriptors)
          {
            final Property property = BeanProperty.create(descriptor.getName());
            
            if (property.isReadable(bean))
              {
                logger.info(">>>> %s.%s=%s", bean, descriptor.getName(), property.getValue(bean));
              }
          }
        
        logger.info("======= dump(%s) done", bean);
      }
    
    /***************************************************************************
     * 
     *
     **************************************************************************/
    private Bean createDecoratorBean (final Bean bean)
      {
        try 
          {
            beanClass.getMethod("addPropertyChangeListener", PropertyChangeListener.class);
            return bean;
          }
        catch (NoSuchMethodException e) 
          {
            return MetadataItemEnhancer.createEnhancedItem(bean);
          }
        catch (SecurityException e) 
          {
            throw new RuntimeException();
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    private void addEventDispatchThreadAssertor (final Component component, 
                                                 final EventDispatchThreadAssertor listener) 
      {
        if (component instanceof JComponent)
          {
            final JComponent jComponent = (JComponent)component;
            jComponent.addPropertyChangeListener(listener);
            jComponent.addVetoableChangeListener(listener);
        
            for (final Component child : jComponent.getComponents())
              {
                addEventDispatchThreadAssertor(child, listener);
              }
          }
      }
    
    /***************************************************************************
     *
     *
     **************************************************************************/
    private static void addPropertyChangeListener (final Object bean, final PropertyChangeListener listener)
      {
        try
          {
            final Method method = bean.getClass().getMethod("addPropertyChangeListener", PropertyChangeListener.class);
            method.invoke(bean, listener);
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
          }
      }
  }
